<?php

namespace WebsiteBundle\Entity;

use Bazookas\CommonBundle\Entity\Base\BaseEntity;
use Doctrine\Common\Collections\ArrayCollection;
use Doctrine\ORM\Mapping as ORM;
use Symfony\Component\Validator\Constraints as Assert;
use Symfony\Component\Validator\Context\ExecutionContextInterface;

/**
 * @ORM\Entity(repositoryClass="WebsiteBundle\Repository\PsychEducationRepository")
 * @ORM\HasLifecycleCallbacks
 */
class PsychEducationPanel extends BaseEntity
{

  //Type constants
  const TYPE_TEXT = 'TEXT';
  const TYPE_RADIO_QUESTION = 'RADIO_QUESTION';
  const TYPE_CHECKBOX_QUESTION = 'CHECKBOX_QUESTION';

  /**
   * @ORM\ManyToOne(targetEntity="Experiment", inversedBy="psychEducationPanels")
   */
  private $experiment;

  /**
   * @ORM\Column(type="string")
   * @var string
   * @Assert\NotBlank
   * @Assert\Choice({"TEXT", "RADIO_QUESTION", "CHECKBOX_QUESTION"})
   */
  private $type;

  /**
   * @ORM\Column(type="string")
   * @var string
   * @Assert\NotBlank
   */
  private $title;

  /**
   * @ORM\Column(type="text", nullable=true)
   * @var string
   */
  private $text = self::TYPE_TEXT;

  /**
   * @ORM\Column(type="string", nullable=true)
   * @var string
   */
  private $question;

  /**
   * @ORM\OneToMany(targetEntity="PsychEducationResponse", mappedBy="panel", cascade="all", orphanRemoval=true)
   * @var PsychEducationResponse[]
   * @Assert\Valid
   */
  private $responses;

  public function __construct()
  {
    parent::__construct();
    $this->responses = new ArrayCollection();
  }

  public function getExperiment()
  {
    return $this->experiment;
  }

  public function setExperiment($experiment)
  {
    $this->experiment = $experiment;

    return $this;
  }

  public function getType()
  {
    return $this->type;
  }

  public function setType($type)
  {
    $this->type = $type;

    return $this;
  }

  public function getTitle()
  {
    return $this->title;
  }

  public function setTitle($title)
  {
    $this->title = $title;

    return $this;
  }

  public function getText()
  {
    return $this->text;
  }

  public function setText($text)
  {
    $this->text = $text;

    return $this;
  }

  public function getQuestion()
  {
    return $this->question;
  }

  public function setQuestion($question)
  {
    $this->question = $question;

    return $this;
  }

  public function getResponses()
  {
    return $this->responses;
  }

  public function setResponses($responses)
  {
    $this->responses = $responses;

    return $this;
  }

  public function addResponse(PsychEducationResponse $response)
  {
    $response->setPanel($this);
    $this->responses->add($response);

    return $this;
  }

  public function removeResponse(PsychEducationResponse $response)
  {
    $this->responses->removeElement($response);

    return $this;
  }

  /**
   * @Assert\Callback
   * @param ExecutionContextInterface $context
   * @param $payload
   */
  public function validate(ExecutionContextInterface $context, $payload)
  {
    if ($this->type !== self::TYPE_TEXT) {
      //Check if a question was added
      if (empty($this->question)) {
        $context->buildViolation('admin.entities.psychEducationPanel.questionRequired')
          ->atPath('question')
          ->addViolation();
      }

      //Check if the responses are correct
      $this->validateResponses($context, $payload);

    }
  }

  /**
   * @param ExecutionContextInterface $context
   * @param $payload
   */
  private function validateResponses(ExecutionContextInterface $context, $payload) {
    $hasCorrectResponse = false;
    foreach ($this->responses as $key => $response) {
      //If its a radio question than we may only have exactly 1 correct answer
      if ($hasCorrectResponse && $this->type === self::TYPE_RADIO_QUESTION && $response->isCorrect()) {
        $context->buildViolation('admin.entities.psychEducationPanel.radioOnlyOneCorrectResponse')
          ->atPath('responses['.$key.'].isCorrect')
          ->addViolation();
      }

      if ($response->isCorrect()) {
        $hasCorrectResponse = true;

        //If its a checkbox question than all we care about is that there is at least 1 correct answer
        if ($this->type === self::TYPE_CHECKBOX_QUESTION) {
          break;
        }
      }
    }

    if (!$hasCorrectResponse) {
      $context->buildViolation('admin.entities.psychEducationPanel.noCorrectResponses')
        ->atPath('isCorrect')
        ->addViolation();
    }
  }

}
