<?php

namespace WebsiteBundle\Repository;

use Bazookas\AdminBundle\Repository\Base\BaseRepository;
use Doctrine\ORM\Query;
use WebsiteBundle\Entity\Experiment;
use WebsiteBundle\Entity\Participant;
use WebsiteBundle\Entity\Quote;
use WebsiteBundle\Entity\Researcher;
use WebsiteBundle\Security\Roles;

class ExperimentRepository extends BaseRepository
{

  public function getEntityClass() {
    return Experiment::class;
  }

  /**
   * Get the currently active experiments
   * @param int $returnType
   * @return array|Experiment[]
   */
  public function getCurrentExperiments($returnType = Query::HYDRATE_ARRAY) {
    $qb = $this->createQueryBuilder('experiment');
    $qb
      ->addSelect('partial participants.{id}')
      ->join('experiment.participants', 'participants', 'WITH', $qb->expr()->eq('participants.active', ':active'))
      ->setParameter('active', true)
    ;

    return $qb->getQuery()->getResult($returnType);
  }

  /**
   * Get all of the experiments that have not started yet
   * @param Researcher $user
   * @param int $returnType
   * @return Experiment[]
   */
  public function getDraftExperimentsForUser(Researcher $user, $returnType = Query::HYDRATE_OBJECT) {
    $qb = $this->createQueryBuilder('experiment');
    $qb
      ->andWhere('SIZE(experiment.sessions) = 0')
      ->leftJoin('experiment.researchers', 'researchers')
    ;

    if (!in_array(Roles::ROLE_ADMIN_RESEARCHER, $user->getRoles())) {
      $qb->andWhere($qb->expr()->eq('researchers.id', $user->getId()));
    }

    return $qb->getQuery()->getResult($returnType);
  }

  /**
   * Get all of the experiments with at least 1 session
   * @param Researcher $user
   * @param int $returnType
   * @return Experiment[]
   */
  public function getCurrentExperimentsForUser(Researcher $user, $returnType = Query::HYDRATE_OBJECT) {
    $qb = $this->createQueryBuilder('experiment');
    $qb
      ->andWhere('SIZE(experiment.sessions) > 0')
      ->leftJoin('experiment.researchers', 'researchers')

      ->andWhere($qb->expr()->neq('experiment.activeParticipantCount', 0))
    ;

    if (!in_array(Roles::ROLE_ADMIN_RESEARCHER, $user->getRoles())) {
      $qb->andWhere($qb->expr()->eq('researchers.id', $user->getId()));
    }

    return $qb->getQuery()->getResult($returnType);
  }

  /**
   * Get all of the experiments for which all of the users are inactive
   * @param Researcher $user
   * @param int $returnType
   * @return Experiment[]
   */
  public function getInactiveExperimentsForUser(Researcher $user, $returnType = Query::HYDRATE_OBJECT) {
    $qb = $this->createQueryBuilder('experiment');
    $qb
      ->andWhere('SIZE(experiment.sessions) > 0')
      ->leftJoin('experiment.researchers', 'researchers')

      ->andWhere($qb->expr()->eq('experiment.activeParticipantCount', 0))
    ;

    if (!in_array(Roles::ROLE_ADMIN_RESEARCHER, $user->getRoles())) {
      $qb->andWhere($qb->expr()->eq('researchers.id', $user->getId()));
    }

    return $qb->getQuery()->getResult($returnType);
  }

  /**
   * @param Experiment $experiment
   */
  public function incrementSessionCount(Experiment $experiment) {
    $qb = $this->entityManager->createQueryBuilder();
    $qb
      ->update(Experiment::class, 'experiment')
      ->set('experiment.sessionCount', 'experiment.sessionCount + 1')
      ->andWhere($qb->expr()->eq('experiment.id', ':id'))
      ->setParameter('id', $experiment->getId())
    ;

    $qb->getQuery()->execute();
  }

  /**
   * @param Researcher $user
   * @param int $returnType
   * @return array
   */
  public function getAllExperimentsForUser($isAdmin, Researcher $user, $returnType = Query::HYDRATE_OBJECT) {
    $qb = $this->createQueryBuilder('experiment');
    $qb
      ->leftJoin('experiment.researchers', 'researchers')
    ;

    if (!$isAdmin) {
      $qb->andWhere($qb->expr()->eq('researchers.id', $user->getId()));
    }

    return $qb->getQuery()->getResult($returnType);
  }

  public function getActiveParticipantCount(Experiment $experiment) {
    $qb = $this->entityManager->createQueryBuilder();
    $qb
      ->select('COUNT(participant.id) as participantCount', 'participant.active')
      ->from(Participant::class, 'participant')
      ->groupBy('participant.active')
      ->andWhere($qb->expr()->eq('participant.experiment', ':experiment'))
      ->setParameter('experiment', $experiment)
    ;

    return $qb->getQuery()->getArrayResult();
  }

  public function getRemainingQuotes(array $unlockedQuotes, Experiment $experiment, $type = Quote::TYPE_GENERIC) {
    $qb = $this->createQueryBuilder('experiment');

    if (empty($unlockedQuotes)) {
      $qb
        ->addSelect('quote')
        ->leftJoin('experiment.availableQuotes', 'quote')
      ;
    } else {
      $qb
        ->addSelect('quote')
        ->leftJoin('experiment.availableQuotes', 'quote', 'WITH', $qb->expr()->notIn('quote.id', ':unlockedQuotes'))
        ->setParameter('unlockedQuotes', $unlockedQuotes)
      ;
    }

    $qb
      ->andWhere($qb->expr()->eq('experiment.id', ':experiment'))
      ->andWhere($qb->expr()->eq('quote.type', ':type'))
      ->setParameter('experiment', $experiment)
      ->setParameter('type', $type)
    ;

    $experiment = $qb->getQuery()->getOneOrNullResult();

    return $experiment->getAvailableQuotes();
  }
}
