<?php

namespace WebsiteBundle\Repository;

use Bazookas\AdminBundle\Repository\Base\BaseRepository;
use Doctrine\ORM\Query;
use WebsiteBundle\Entity\Experiment;
use WebsiteBundle\Entity\Participant;
use WebsiteBundle\Entity\TestAnswer;
use WebsiteBundle\Entity\TestSession;

class TestSessionRepository extends BaseRepository
{

  public function getEntityClass() {
    return TestSession::class;
  }

  /**
   * @param Participant $participant
   * @param Experiment $experiment
   * @param int $returnType
   * @return TestSession|null
   */
  public function getNextSession(Participant $participant, Experiment $experiment, $returnType = Query::HYDRATE_ARRAY) {
    $qb = $this->createQueryBuilder('session');
    $qb
      ->andWhere($qb->expr()->eq('session.participant', ':participant'))
      ->andWhere($qb->expr()->eq('session.experiment', ':experiment'))
      ->andWhere($qb->expr()->isNull('session.completed'))
      ->setParameters([
        'participant' => $participant,
        'experiment' => $experiment
      ])
      ->addOrderBy('session.id', 'DESC')
      ->setMaxResults(1)
    ;

    $results = $qb->getQuery()->getResult($returnType);
    return !empty($results) ? $results[0] : null;
  }

  /**
   * @param Participant $participant
   * @param Experiment $experiment
   * @return TestSession|null
   */
  public function getLastSession(Participant $participant, Experiment $experiment) {
    $qb = $this->createQueryBuilder('session');
    $qb
      ->andWhere($qb->expr()->eq('session.participant', ':participant'))
      ->andWhere($qb->expr()->eq('session.experiment', ':experiment'))
      ->andWhere($qb->expr()->isNotNull('session.completed'))
      ->setParameters([
        'participant' => $participant,
        'experiment' => $experiment
      ])
      ->addOrderBy('session.completed', 'DESC')
      ->setMaxResults(1)
    ;

    return $qb->getQuery()->getOneOrNullResult();
  }

  /**
   * @param Participant $participant
   * @param TestSession $session
   * @return array
   */
  public function getFirstSessionResults(Participant $participant, TestSession $session) {
    $qb = $this->entityManager->createQueryBuilder();
    $qb
      ->select($qb->expr()->count('ans.id'), 'ans.interval')
      ->from(TestAnswer::class, 'ans')
      ->andWhere($qb->expr()->eq('ans.correct', ':correct'))
      ->andWhere($qb->expr()->eq('ans.participant', ':participant'))
      ->andWhere($qb->expr()->eq('ans.session', ':session'))
      ->groupBy('ans.interval')

      ->setParameters([
        'correct' => true,
        'participant' => $participant,
        'session' => $session
      ])
    ;

    $results = $qb->getQuery()->getArrayResult();
    return $results;
  }

  /**
   * @param Participant $participant
   * @return TestSession[]
   */
  public function getSessions(Participant $participant) {
    $qb = $this->createQueryBuilder('session');
    $qb
      ->andWhere($qb->expr()->eq('session.participant', ':participant'))
      ->andWhere($qb->expr()->isNotNull('session.completed'))
      ->setParameter('participant', $participant)
      ->orderBy('session.number', 'ASC')
    ;

    return $qb->getQuery()->getArrayResult();
  }

  /**
   * @param Participant $participant
   * @param Experiment $experiment
   * @return null|TestSession
   */
  public function getLastSessionWithoutMoodAnswers(Participant $participant, Experiment $experiment) {
    $qb = $this->createQueryBuilder('session');
    $qb
      ->andWhere($qb->expr()->eq('session.participant', ':participant'))
      ->andWhere($qb->expr()->eq('session.experiment', ':experiment'))
      ->andWhere($qb->expr()->isNotNull('session.completed'))
      ->andWhere('SIZE(session.moodAnswers) = 0')
      ->andWhere($qb->expr()->neq('session.number', 1))
      ->setParameters([
        'participant' => $participant,
        'experiment' => $experiment
      ])
      ->addOrderBy('session.completed', 'DESC')
      ->setMaxResults(1)
    ;

    return $qb->getQuery()->getOneOrNullResult();
  }

  /**
   * Get the participants whom need to be sent a reminder email
   * @param int $experimentId
   * @param int $time
   * @param string $type 'before' => people who need reminding before the scheduled time, 'after' => people who need
   * reminding after the scheduled time.
   * @param int $returnType
   * @return array|TestSession[]
   */
  public function getParticipantsThatNeedReminding($experimentId, $time, $type = 'before', $returnType = Query::HYDRATE_OBJECT) {
    $reminderTime = new \DateTime('now', new \DateTimeZone('UTC'));
    if ($type == 'before') {
      $reminderTime->modify('+'.$time.' minutes');
    } else {
      $reminderTime->modify('-'.$time.' minutes');
    }

    $qb = $this->_em->createQueryBuilder();
    $qb
      ->select('session')
      ->from(TestSession::class, 'session')
      ->andWhere($qb->expr()->isNull('session.completed'))
      ->andWhere($qb->expr()->lt('session.scheduled', ':reminderTime'))
      ->join('session.experiment', 'experiment', 'WITH', $qb->expr()->eq('experiment.id', ':experiment'))
    ;

    if ($type == 'before') {
      $qb
        ->andWhere($qb->expr()->eq('session.sentBeforeNotification', ':notified'))
        ->andWhere($qb->expr()->neq('session.number', 1))
      ;
    } else {
      $qb->andWhere($qb->expr()->eq('session.sentAfterNotification', ':notified'));
    }

    $qb
      ->addSelect('participant')
      ->join('session.participant', 'participant', 'WITH', $qb->expr()->eq('participant.active', ':active'))

      ->setParameters([
        'reminderTime' => $reminderTime,
        'notified' => false,
        'active' => true,
        'experiment' => $experimentId
      ])
    ;

    return $qb->getQuery()->getResult($returnType);
  }

  public function createListQueryBuilder(array $fields, array $params) {
    if (!empty($params)) {
      $qb = $this->createQueryBuilder('e');
      $qb
        ->andWhere($qb->expr()->eq('e.participant', ':participant'))
        ->setParameter('participant', $params['participantId'])
      ;

      return $qb;
    } else {
      return parent::createListQueryBuilder($fields, $params);
    }
  }

  /**
   * @param $participantId
   * @return \Doctrine\ORM\QueryBuilder
   */
  public function createParticipantExportQueryBuilder($participantId) {
    $qb = $this->createQueryBuilder('session');

    $qb
      ->addSelect('participant')
      ->leftJoin('session.participant', 'participant')

      ->addSelect('moodAnswers')
      ->leftJoin('session.moodAnswers', 'moodAnswers')

      ->addSelect('question')
      ->leftJoin('moodAnswers.question', 'question')

      ->addSelect('partial experiment.{id}')
      ->leftJoin('session.experiment', 'experiment')

      ->andWhere($qb->expr()->eq('session.participant', ':participant'))
      ->andWhere($qb->expr()->isNotNull('session.completed'))
      ->setParameter('participant', $participantId)
    ;

    return $qb;
  }

  /**
   * @param int $sessionId
   * @return int
   */
  public function getParticipantId($sessionId) {
    $qb = $this->createQueryBuilder('session');
    $qb
      ->select('IDENTITY(session.participant)')
      ->andWhere($qb->expr()->eq('session.id', ':sessionId'))
      ->setParameter('sessionId', $sessionId)
    ;

    return $qb->getQuery()->getSingleScalarResult();
  }

}
