<?php

namespace Bazookas\AdminBundle\Controller\Base;

use Bazookas\AdminBundle\AdminElements\Elements\Actions\ListActions\EditBulkActionElement;
use Bazookas\AdminBundle\AdminElements\Elements\Actions\ListActions\EditRowActionElement;
use Bazookas\AdminBundle\AdminElements\Elements\Actions\ListActions\PickRowActionElement;
use Bazookas\AdminBundle\AdminElements\Elements\Actions\ListActions\RemoveBulkActionElement;
use Bazookas\AdminBundle\AdminElements\Elements\Actions\ListActions\RemoveRowActionElement;
use Bazookas\AdminBundle\AdminElements\Elements\Actions\ListActions\ViewRowActionElement;
use Bazookas\AdminBundle\AdminElements\Elements\Actions\PageActions\AddPageActionElement;
use Bazookas\AdminBundle\AdminElements\Elements\Actions\PageActions\CancelPageActionElement;
use Bazookas\AdminBundle\AdminElements\Elements\Actions\PageActions\SubmitPageActionElement;
use Bazookas\AdminBundle\PageBuilder\Base\BaseFormPageBuilder;
use Bazookas\AdminBundle\PageBuilder\BulkEditPageBuilder;
use Bazookas\AdminBundle\PageBuilder\Interfaces\ListPageBuilderInterface;
use Bazookas\AdminBundle\PageBuilder\ListPageBuilder;
use Bazookas\AdminBundle\Repository\Interfaces\ListRepositoryInterface;
use Bazookas\CommonBundle\Entity\Interfaces\AccessControlInterface;
use Bazookas\CommonBundle\Exception\EntityAccessDeniedException;
use Bazookas\CommonBundle\Exception\LinkedEntityException;
use Bazookas\CommonBundle\Exception\NotImplementedException;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Exception\AccessDeniedHttpException;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;

abstract class BaseAdminListController extends BaseAdminActionController
{

  public function __construct() {
    parent::__construct();

    //Define the default page builders here, entire list shown for completeness
    $this->builders = [
      self::ACTION_LIST => ListPageBuilder::class,
      self::ACTION_ADD => BaseFormPageBuilder::class,
      self::ACTION_EDIT => BaseFormPageBuilder::class,
      self::ACTION_REMOVE => null,
      self::ACTION_VIEW => null,
      self::ACTION_BULK_EDIT => BulkEditPageBuilder::class,
      self::ACTION_BULK_REMOVE => null,

      self::AJAX_LIST => ListPageBuilder::class
    ];
  }

  /**
   * @inheritDoc
   */
  protected function hasAccess($action) {
    switch($action) {
      case self::ACTION_LIST:
      case self::ACTION_EDIT:
      case self::ACTION_ADD:
      case self::ACTION_REMOVE:
        return true;
      default:
        return false;
    }
  }

  /**
   * Get the short class name of the entity.
   * @return string
   * @throws \Exception
   */
  public function getEntityShortClassName() {
    $entityClass = $this->getEntityClass();
    if (!empty($entityClass)) {
      $refClass = new \ReflectionClass($this->getEntityClass());
      return lcfirst($refClass->getShortName());
    } else {
      throw new \Exception('No entity class defined. If your controller does not make use of an entity then you must overwrite the \'getEntityShortClassName\' method.');
    }
  }

  /* ------------------------------------------------------------------------------
   *                              MODIFY BUILDERS
     ------------------------------------------------------------------------------ */

  /**
   * @param Request $request
   * @param ListPageBuilderInterface $builder
   * @return ListPageBuilderInterface
   */
  protected function modifyListBuilder(Request $request, ListPageBuilderInterface $builder) {
    //Add the list actions
    $actions = [
      self::ACTION_VIEW,
      self::ACTION_ADD,
      self::ACTION_EDIT,
      self::ACTION_REMOVE,
      self::ACTION_BULK_EDIT,
      self::ACTION_BULK_REMOVE
    ];

    $route = $request->get('_route');
    $isPicker = $request->get('picker');
    if ($isPicker) {
      $builder->addAction(new PickRowActionElement([], []));
    }
    else {
      foreach($actions as $action) {
        if ($this->hasAccess($action)) {
          $actionItem = null;
          switch($action) {
            case self::ACTION_VIEW:
              $builder->addAction(new ViewRowActionElement([], ['route' => $route ]));
              break;
            case self::ACTION_EDIT:
              $builder->addAction(new EditRowActionElement([], ['route' => $route ]));
              break;
            case self::ACTION_REMOVE:
              $builder->addAction(new RemoveRowActionElement([], [ 'route' => $route ]));
              break;
            case self::ACTION_BULK_EDIT:
              $builder->addAction(new EditBulkActionElement([], [ 'route' => $route ]));
              break;
            case self::ACTION_BULK_REMOVE:
              $builder->addAction(new RemoveBulkActionElement([], [ 'route' => $route ]));
              break;
          }
        }
      }
    }

    //If the user is allowed to add items, then add a page action for the add page (unless we're in a picker form)
    if ( ($this->hasAccess(self::ACTION_ADD)) && (!$isPicker)) {
      $builder->addAction(new AddPageActionElement([], [ 'route' => $route ]));
    }

    return $builder;
  }

  /**
   * @param Request $request
   * @param BaseFormPageBuilder $builder
   * @return BaseFormPageBuilder
   * @throws \Exception|AccessDeniedHttpException
   */
  protected function modifyAddBuilder(Request $request, BaseFormPageBuilder $builder) {
    $entity = $builder->getEntity();

    //Perform access control check (if necessary)
    if ($entity instanceof AccessControlInterface) {
      $auth = $this->get('security.authorization_checker');
      if (!$entity->canAdd($auth, $this->getUser())) {
        throw new AccessDeniedHttpException();
      }
    }

    //Get the fully qualified class name of the form we should be using
    $formClass = $this->getFormClass(self::ACTION_ADD);
    if (empty($formClass)) {
      throw new \Exception('No form class defined for action ' . self::ACTION_ADD);
    }

    //Create the form
    $form = $this->createForm($formClass, $entity, $builder->getFormOptions());

    //Add the elements to the builder
    $builder
      ->setForm($form)
      ->setHeaderLabel('admin.entities.' . lcfirst($this->getEntityShortClassName()) . '.addLabel')
    ;

    //Add the page actions
    $route = $request->get('_route');
    $builder
      ->addAction(new CancelPageActionElement([], [
        'route' => $route,
        'action' => null
      ]))
      ->addAction(new SubmitPageActionElement([], [
        'route' => $route,
        'action' => $request->get('_action'),
        'iconClass' => 'ti-save'
      ]))
    ;
    return $builder;
  }

  /**
   * @param Request $request
   * @param BaseFormPageBuilder $builder
   * @return BaseFormPageBuilder
   * @throws \Exception|AccessDeniedHttpException
   */
  protected function modifyEditBuilder(Request $request, BaseFormPageBuilder $builder) {
    //Set the form data
    $builder->setEntityId($request->get('_id'));

    $entity = $builder->getEntity();

    //Perform access control check (if necessary)
    if ($entity instanceof AccessControlInterface) {
      $auth = $this->get('security.authorization_checker');
      if (!$entity->canEdit($auth, $this->getUser())) {
        throw new AccessDeniedHttpException();
      }
    }

    //Get the fully qualified class name of the form we should be using
    $formClass = $this->getFormClass(self::ACTION_EDIT);
    if (empty($formClass)) {
      throw new \Exception('No form class defined for action ' . self::ACTION_EDIT);
    }

    //Create the form
    $form = $this->createForm($formClass, $entity, $builder->getFormOptions());

    //Add the elements to the builder
    $builder
      ->setForm($form)
      ->setHeaderLabel('admin.entities.' . lcfirst($builder->getEntityShortClassName()) . '.editLabel')
    ;

    //Add the page actions
    $route = $request->get('_route');
    $builder
      ->addAction(new CancelPageActionElement([], [
        'route' => $route,
        'action' => null
      ]))
      ->addAction(new SubmitPageActionElement([], [
        'route' => $route,
        'action' => $request->get('_action'),
        'iconClass' => 'ti-save'
      ]))
    ;
    return $builder;
  }

  /**
   * @param Request $request
   * @param BulkEditPageBuilder $builder
   * @return BulkEditPageBuilder
   * @throws \Exception|EntityAccessDeniedException
   */
  protected function modifyBulkEditBuilder(Request $request, BulkEditPageBuilder $builder) {
    //Get the entities we'll be editing
    $entityIds = $request->get('ids');

    if (empty($entityIds)) {
      throw new EntityAccessDeniedException('Access denied');
    }

    $builder->setEntityIds($entityIds);
    $entities = $builder->getEntities();

    //Perform access control check (if necessary)
    $auth = $this->get('security.authorization_checker');
    foreach($entities as $entity) {
      if ($entity instanceof AccessControlInterface) {
        if (!$entity->performAccessCheck($auth, $this->getUser(), self::ACTION_BULK_EDIT)) {
          throw new EntityAccessDeniedException('Access denied', 0, null, $entity);
        }
      }
    }

    //Set the form data by create a single empty instance of the entity
    $entity = $builder->getEntity();

    //Get the fully qualified class name of the form we should be using
    $formClass = $this->getFormClass(self::ACTION_BULK_EDIT);
    if (empty($formClass)) {
      throw new \Exception('No form class defined for action ' . self::ACTION_EDIT);
    }

    //Create the form
    $form = $this->createForm($formClass, $entity, $builder->getFormOptions());

    //Add the elements to the builder
    $builder
      ->setForm($form)
    ;

    //Add the page actions
    $route = $request->get('_route');
    $builder
      ->addAction(new CancelPageActionElement([], [
        'route' => $route,
        'action' => null
      ]))
      ->addAction(new SubmitPageActionElement([], [
        'route' => $route,
        'action' => self::ACTION_BULK_EDIT,
        'iconClass' => 'ti-plus'
      ]))
    ;
    return $builder;
  }

  /* ------------------------------------------------------------------------------
 *                              ACTIONS
   ------------------------------------------------------------------------------ */
  /**
   * @param Request $request
   * @param string $action
   * @return \Symfony\Component\HttpFoundation\Response
   */
  protected function addAction(Request $request, $action) {
    if ($request->getMethod() === Request::METHOD_GET) {
      return $this->defaultAction($request, $action);
    } else {
      return $this->formAction($request, $action);
    }
  }

  /**
   * @param Request $request
   * @param string $action
   * @return \Symfony\Component\HttpFoundation\Response
   */
  protected function editAction(Request $request, $action) {
    if ($request->getMethod() === Request::METHOD_GET) {
      return $this->defaultAction($request, $action);
    } else {
      return $this->formAction($request, $action);
    }
  }

  /**
   * @param Request $request
   * @param string $action
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   * @throws \Exception
   */
  protected function removeAction(Request $request, $action) {
    //Get the entity
    $entityId = $request->get('_id');
    $entityRepo = $this->get('doctrine.orm.entity_manager')->getRepository($this->getEntityClass());
    $entity = $entityRepo->find($entityId);

    //If the entity was not found throw an error
    if (empty($entity)) {
      return $this->accessDeniedRedirect($request);
    }

    //Perform access control check (if necessary)
    if ($entity instanceof AccessControlInterface) {
      $auth = $this->get('security.authorization_checker');
      if (!$entity->canDelete($auth, $this->getUser())) {
        return $this->accessDeniedRedirect($request);
      }
    }

    //Check if the entity repository implements the necessary interface
    if (!$entityRepo instanceof ListRepositoryInterface) {
      throw new \Exception('The repository class must implement the interface: ' . ListRepositoryInterface::class);
    }

    //Remove the entities
    $translator = $this->get('translator');
    try {
      $entityRepo->removeEntities([$entity]);

      //Add a flash message

      $messageKey = $this->getActionFlashMessageKey($action, $this->getEntityShortClassName());
      $this->addFlash('success',
        $translator->trans($messageKey, [], 'admin')
      );
    } catch (LinkedEntityException $e) {
      $linkedTo = [];
      foreach($e->getPayload() as $entityName => $items) {
        foreach($items as $item) {
          $entityName = $translator->trans('admin.entities.'.lcfirst($this->getEntityShortClassName()). '.name', [], 'admin');
          $message = $translator->trans('admin.list.messages.couldNotBeRemoved', array(
            '%entity%' => $entityName
          ), 'admin');

          if (method_exists($item,'toHumanReadable')) {
            $linkedTo[$message][] = $item->toHumanReadable();
          }
        }
      }

      $this->addFlash(
        'error', $linkedTo
      );
    }

    return $this->redirectToRoute($request->get('_route'));
  }

  /**
   * @param Request $request
   * @param string $action
   * @return \Symfony\Component\HttpFoundation\Response
   */
  protected function bulkEditAction(Request $request, $action) {
    if ($request->getMethod() === Request::METHOD_GET) {
      return $this->defaultAction($request, $action);
    } else {
      return $this->formAction($request, $action);
    }
  }

  /**
   * @param Request $request
   * @param $action
   * @return \Symfony\Component\HttpFoundation\Response
   */
  protected function bulkRemoveAction(Request $request, $action) {

    //TODO implement this
    throw new NotImplementedException();

  }

  /**
   * @param Request $request
   * @param $action
   * @return \Symfony\Component\HttpFoundation\Response
   * @throws \Exception
   */
  public function ajaxListAction(Request $request, $action) {
    //Get the page builder for the current action
    $builder = $this->getPageBuilder($action);

    //Check if a page builder was defined
    if (empty($builder)) {
      throw new NotFoundHttpException();
    }

    if ($builder instanceof ListPageBuilderInterface) {
      //Allow the builder to be modified using the modify hook
      $builder = $this->modifyBuilder($request, self::ACTION_LIST, $builder);

      //Build and add the element
      $listElement = $builder->buildListTable($request);
      $this->elements[] = $listElement;

      //Get any custom template data
      $data = $builder->getTemplateData();
    } else if (is_callable($builder)){
      $data = $builder();
    } else {
      throw new NotFoundHttpException();
    }

    //Merge the custom template data with the default template data
    $templateData  = $this->getTemplateData($request, $action, $data);

    //Generate the response
    $this->layoutTemplate = self::TEMPLATE_LAYOUT_CONTENT_ONLY;
    return $this->createResponse($request, $templateData, $action);
  }

  /**
   * @param $action
   * @return string the fully qualified class name of the form
   */
  protected abstract function getFormClass($action);
}
