<?php

namespace Bazookas\AdminBundle\EventListener\Entity;

use Bazookas\AdminBundle\Entity\Interfaces\AuditableInterface;
use Bazookas\AdminBundle\Entity\Interfaces\AuditLogInterface;
use Bazookas\AdminBundle\Entity\Interfaces\Versionable;
use Doctrine\Common\Cache\Version;
use Doctrine\ORM\EntityManager;
use Doctrine\ORM\Event\LifecycleEventArgs;
use Doctrine\ORM\Event\OnFlushEventArgs;
use Doctrine\ORM\Mapping\ClassMetadata;
use Doctrine\ORM\PersistentCollection;
use Doctrine\ORM\UnitOfWork;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorage;

class VersioningListener {

  /**
   * @var EntityManager
   */
  private $em;

  private $tokenStorage;

  public function __construct(TokenStorage $tokenStorage) {
    $this->tokenStorage = $tokenStorage;
  }

  public function prePersist(LifecycleEventArgs $args) {
    $object = $args->getObject();
    if ($object instanceof Versionable) {
      $token = $this->tokenStorage->getToken();
      $user = null;
      if (!empty($token)) {
        $user = $token->getUser();
      }

      $object->setCreator($user);
    }
  }

  public function onFlush(OnFlushEventArgs $args) {
    $this->em = $args->getEntityManager();
    $uow = $this->em->getUnitOfWork();

    foreach ($uow->getScheduledEntityUpdates() as $entity) {
      if ($entity instanceof Versionable) {
        $changeSet = $uow->getEntityChangeSet($entity);
        $metaData = $this->em->getClassMetadata(get_class($entity));
        $newVersion = $this->createNewVersion($changeSet, $entity, $metaData);

        $this->em->persist($newVersion);
        $uow->computeChangeSet($this->em->getClassMetadata(get_class($newVersion)), $newVersion);
        $uow->computeChangeSet($this->em->getClassMetadata(get_class($entity)), $entity);
      }
    }
  }

  private function createNewVersion(array $changeSet, Versionable $entity, ClassMetadata $metaData) {
    $token = $this->tokenStorage->getToken();
    $user = null;
    if (!empty($token)) {
      $user = $token->getUser();
    }

    $associations = $this->filterAssociations($metaData->getAssociationMappings());

    $newVersion = clone $entity;
    $newVersion
      ->setCreated(new \DateTime('now', new \DateTimeZone('UTC')))
      ->setId(null)
      ->setChildren([])
      ->setCreator($user)
      ->setParent($entity)
    ;

    // clone ManyToMany relations that are not owned by this entity
    $this->cloneRelations($entity, $newVersion, $associations);

    //Fill the entity with the old values again
    foreach($changeSet as $field => $data) {
      $entity->{'set'.ucfirst($field)}($data[0]);
    }

    $entity
      ->addChild($newVersion)
    ;

    return $newVersion;
  }

  private function filterAssociations(array $associations) {
    $relations = array();

    // filter out the default versionable relations
    unset($associations['children']);
    unset($associations['parent']);
    unset($associations['creator']);

    // filter out OneToOne and ManyToOne relations as they clone just fine
    foreach ($associations as $key => $association) {
      if ( ($association['type'] == 1) || ($association['type'] == 2) || ($association['type'] == 4) ) {
        unset($associations[$key]);
      } elseif($association['type'] == 8) {
        if (is_null($association['mappedBy'])) {
          // entity owns the relation, so with cloning it will clone the relation
          unset($associations[$key]);
        } else {
          $relations[$key] = $association['mappedBy'];
        }
      }
    }

    return $relations;
  }

  private function cloneRelations($oldVersion, $newVersion, $associations) {
    foreach ($associations as $field => $mapping) {
      /** @var PersistentCollection $entities */
      $entities = $oldVersion->{'get'.ucfirst($field)}();
      $entities->initialize();
      $entities = $entities->unwrap();

      foreach ($entities as $entity) {
        /** @var PersistentCollection $oldFields */
        $oldFields = $entity->{'get'.ucfirst($mapping)}();
        $oldFields->initialize();
        $oldFields = $oldFields->unwrap();
        $oldFields->add($newVersion);
        $entity->{'set'.ucfirst($mapping)}($oldFields);
        $this->em->persist($entity);

        $uow = $this->em->getUnitOfWork();
        $uow->computeChangeSet($this->em->getClassMetadata(get_class($entity)), $entity);
      }
    }
  }
}
