<?php

namespace Bazookas\AdminBundle\Filter;

use Bazookas\AdminBundle\Form\ListFilterCollectionForm;
use Bazookas\CommonBundle\Exception\NotImplementedException;
use Doctrine\ORM\QueryBuilder;
use Symfony\Component\Form\FormFactory;
use Symfony\Component\HttpFoundation\Request;

class ListFilterService
{

  /**
   * @var ListFilter[]
   */
  protected $filters;

  /**
   * @var bool
   */
  protected $maintainFilterState = false;

  public function __construct()
  {
    $this->filters = [];
  }

  /**
   * @return array|ListFilter[]
   */
  public function getFilters()
  {
    return $this->filters;
  }

  /**
   * @param string $fieldName
   * @param ListFilter $filter
   * @return $this
   */
  public function addFilter($fieldName, ListFilter $filter)
  {
    $this->filters[$fieldName] = $filter;

    return $this;
  }

  /**
   * @param string $fieldName
   * @return bool
   */
  protected function hasFilter($fieldName)
  {
    return isset($this->filters[$fieldName]);
  }

  /**
   * @param string $fieldName
   * @return ListFilter|null
   */
  public function getFilter($fieldName)
  {
    if ($this->hasFilter($fieldName)) {
      return $this->filters[$fieldName];
    }

    return null;
  }

  /**
   * @return bool
   */
  public function getMaintainFilterState()
  {
    return $this->maintainFilterState;
  }

  /**
   * @param bool $maintainFilterState
   * @return $this
   */
  public function setMaintainFilterState($maintainFilterState)
  {
    $this->maintainFilterState = $maintainFilterState;

    return $this;
  }

  /**
   * This function parses the filters from the request object
   * @param Request $request
   *
   * @return $this
   */
  public function parseFilters(Request $request)
  {
    $route = $request->get('_route');
    $action = $request->get('_action');
    $session = $request->getSession();
    $filterHash = md5($route . $action);

    $clear = $request->query->get('clearFilters');

    if (!isset($clear)) {
      //handle any other filters passed in the url
      $filters = $request->query->get('filter', []);
      $filterForm = $request->query->get('filters', []);

      //clear any existing filters
      $this->filters = [];

      if (!empty($filterForm)) {
        $filters = $filterForm;
      }

      foreach ($filters as $filter) {
        $filterObject = new ListFilter();
        $filterObject->setField($filter['field']);
        $filterObject->setOperator($filter['operator']);
        if (isset($filter['value'])) {
          $filterObject->setValue($filter['value']);
        } else {
          $filterObject->setValue(0);
        }

        if (isset($filter['type'])) {
          $filterObject->setType($filter['type']);
        }

        $this->filters[] = $filterObject;
      }

      if (empty($this->filters) && $this->getMaintainFilterState()) {
        $sessionFilters = $session->get('filters');

        if (isset($sessionFilters[$filterHash])) {
          $this->filters = $sessionFilters[$filterHash];
          return $this;
        } else if (!empty($sessionFilters)) {
          $session->remove('filters');
        }
      }

      //overwrite the session storage, this will remove any other filters still present
      $session->set('filters', [$filterHash => $this->filters]);
    } else {
      $this->filters = [];
      $session->set('filters', [$filterHash => $this->filters]);
    }

    return $this;
  }

  /**
   * This function executes the filters
   * @param queryBuilder $queryBuilder the querybuilder instance to apply filters to
   *
   * @return $this
   */
  public function executeFilters(QueryBuilder &$queryBuilder)
  {
    $filterParameter = 1;
    foreach ($this->filters as &$filter) {

      if (is_object($filter)) {
        $field = $this->joinAndGetFieldName($queryBuilder, $filter->getField());
        $operator = $filter->getOperator();
        $value = $filter->getValue();
        $type = $filter->getType();
      } else {
        $field = $this->joinAndGetFieldName($queryBuilder, $filter['field']);
        $operator = $filter['operator'];
        $value = $filter['value'];
        $type = isset($filter['type']) ? $filter['type'] : ListFilterField::TYPE_TEXT;
      }

      $queryForEmpty = false;
      if (!is_int($value) && empty($value)) {
        $queryForEmpty = true;
        $queryBuilder->andWhere(
          $queryBuilder->expr()->orX(
            $queryBuilder->expr()->{$operator}($field, ":filter$filterParameter"),
            $queryBuilder->expr()->isNull($field)
          )
        );
        $queryBuilder->setParameter(":filter$filterParameter", $value);
      }
      else if ($type === ListFilterField::TYPE_DATETIME) {
        $value = new \DateTime($value, new \DateTimeZone('Europe/Brussels'));
      }

      if ($operator === 'like' || $operator === 'notLike') {
        $value = "%$value%";
      }

      if (!$queryForEmpty) {
        $queryBuilder->andWhere(
          $queryBuilder->expr()->{$operator}($field, ":filter$filterParameter")
        );
        $queryBuilder->setParameter(":filter$filterParameter", $value);
      }

      $filterParameter++;
    }

    return $this;
  }

  /**
   * @param QueryBuilder $queryBuilder
   * @param string $fieldPath
   * @return string
   */
  private function joinAndGetFieldName(QueryBuilder $queryBuilder, $fieldPath)
  {
    $rootAliases = $queryBuilder->getRootAliases();

    if (count($rootAliases) > 1) {
      throw new NotImplementedException();
    } else {
      $rootAlias = $rootAliases[0];
    }

    $exploded = explode('.', $fieldPath);
    if (count($exploded) === 1) {
      return $rootAlias.'.' . $fieldPath;
    } else {
      $lastField = array_pop($exploded);
      $previousField = $rootAlias;
      foreach ($exploded as $field) {
        if (!$this->joinExsists($queryBuilder, $previousField, $field)) {
          // add the join to the queryBuilder
          $queryBuilder->leftJoin("$previousField.$field", $field);

          // add the joined entity to the select statement
          // to make sure the joined entities stay filtered
          $queryBuilder->addSelect($field);
        }
        $previousField = $field;
      }

      return "$field.$lastField";
    }
  }

  /**
   * @param QueryBuilder $queryBuilder
   * @param string $prefix
   * @param string $alias
   * @return bool
   */
  private function joinExsists(QueryBuilder $queryBuilder, $prefix, $alias)
  {
    $parts = $queryBuilder->getDQLPart('join');
    if (isset($parts[$prefix])) {
      foreach ($parts[$prefix] as $existingJoin) {
        if ($existingJoin->getAlias() === $alias) {
          return true;
        }
      }
    }

    return false;
  }

  /**
   * @param FormFactory $formFactory
   * @param array $filterFields
   * @return null|\Symfony\Component\Form\FormInterface
   */
  public function createFilterForm(FormFactory $formFactory, array $filterFields) {
    if (!empty($filterFields)) {
      $filters = $this->getFilters();

      if (empty($filters)) {
        $listFilter = new ListFilter();
        $listFilter
          ->setFilterFields($filterFields)
        ;
        $filters = [$listFilter];
      }
      else {
        foreach ($filters as $filter) {
          $filter->setFilterFields($filterFields);
        }
      }

      $filterCollection = new ListFilterCollection();
      $filterCollection->setFilters($filters);

      return $formFactory
        ->createNamedBuilder(null, ListFilterCollectionForm::class, $filterCollection, array(
          'method' => 'GET',
        ))
        ->getForm()
      ;
    }

    return null;
  }
}
