## Modifying page builders

The page builder system allows you to define page builders for each of your actions. If you've followed along with the [first getting started document](./basicListController.md) then you should have a controller that looks similar to:

```php
<?php

namespace WebsiteBundle\Controller\Admin;

use Bazookas\AdminBundle\Controller\Base\BaseAdminListController;
use Bazookas\AdminBundle\PageBuilder\ListPageBuilder;
use Symfony\Component\HttpFoundation\Request;
use WebsiteBundle\Entity\Test;
use WebsiteBundle\Form\Admin\TestAdminType;
use WebsiteBundle\Security\Roles;

class TestAdminController extends BaseAdminListController
{

  protected function getEntityClass()
  {
    return Test::class;
  }

  protected function modifyListBuilder(Request $request, ListPageBuilder $builder)
  {
    $builder = parent::modifyListBuilder($request, $builder);

    $builder
      ->addField('name')
      ->addField('type', [
        'template' => 'WebsiteBundle:List\Fields:testType.html.twig'
      ])
    ;
    
    return $builder;
  }

  protected function getFormClass($action) {
    switch($action) {
      case self::ACTION_ADD:
      case self::ACTION_EDIT:
        return TestAdminType::class;
      default:
        return null;
    }
  }

  protected function hasAccess($action) {
    switch($action) {
      case self::ACTION_LIST:
      case self::ACTION_EDIT:
        return $this->isGranted(Roles::ROLE_ADMIN);
      case self::ACTION_ADD:
        return $this->isGranted(Roles::ROLE_SUPER_ADMIN);
      default:
        return false;
    }
  }

}
```
 
### Creating an add or edit page
 
The admin pages, by default, make use of the page builder system. This allows you to override key sections of an admin page's functionality while letting the CMS system do the heavy lifting. Let's assume you want to create an add or edit page. In order to achieve a basic add or edit page all you have to do is allow access via the `hasAccess` function and to define a form class via the `getFormClass` function. 
 
 However the modify page builder system allows you to do so much more.
 
 #### Adding form options
 
 In order to add form options simply override the `modifyBuilder` function of your action as follows:
 
 ```php
   protected function modifyEditBuilder(Request $request, BaseFormPageBuilder $builder)
   {
     $formOptions = $this->getFormOptions();
     $builder->setFormOptions($formOptions);
     return parent::modifyEditBuilder($request, $builder);
   }
```
 
 #### Using a custom form builder
 
 You can override which builder class is used to render the page in the constructor of your admin list controller:
 
 ```php
 public function __construct() {
     parent::__construct();
 
     $this->builders[BaseAdminListController::ACTION_ADD] = '@WebsiteBundle.PageBuilder.ParticipantFormPageBuilder';
     $this->builders[BaseAdminListController::ACTION_EDIT] = BaseFormBuilder::Class;
     $this->builders[BaseAdminListController::ACTION_VIEW] = ParticipantViewPageBuilder::class;
 }
```
 
 As you can see in the example above the builder array is an associative array containing `<page action> => <page builder>` pairs. The page builder can be defined as a fully qualified class constant such as `ParticipantViewPageBuilder::class` or it can be defined as a service by prodividing a string of the format `@<my service id>`. Note that a callable can also be used.
 
 In the example you can see we mixed the use of a service, a base cms builder (`BaseFormBuilder`) and our own custom page builder class `ParticipantViewPageBuilder`.
 
 The view page builder example might be something as simple as including a second list table for an associated entity:
 
 ```php
 <?php
 
 namespace WebsiteBundle\PageBuilder;
 
 use Bazookas\AdminBundle\PageBuilder\ListPageBuilder;
 use WebsiteBundle\Entity\TestSession;
 
 class ParticipantViewPageBuilder extends ListPageBuilder
 {
   public function getEntityClass()
   {
     return TestSession::class;
   }
 
   /**
    * @inheritDoc
    */
   public function getTemplateData()
   {
     return [];
   }
 }
```
 
### Creating a bulk edit page
In order to enable the bulk edit functionality modify the `hasAccss` function to allow access for the `self::ACTION_BULK_EDIT` action:
 
 ```php
protected function hasAccess($action) {
 switch($action) {
   case self::ACTION_BULK_EDIT:
     return true;
   default:
     return false;
 }
}
```

 Then define a class for use with the bulk edit functionality:
 ```php
protected function getFormClass($action) {
 switch($action) {
   case self::ACTION_BULK_EDIT:
     return ParticipantBulkEditAdminType::class;
   default:
     return null;
 }
}
```

As with all page builders it is possible to modify the bulk edit page builder using the `modifyBuilder` functionality:

```php
  protected function modifyBulkEditBuilder(Request $request, BulkEditPageBuilder $builder)
  {
    $formOptions = $this->getFormOptions();
    $builder->setFormOptions($formOptions);
    return parent::modifyBulkEditBuilder($request, $builder);
  }
```

### Adding custom page actions

When modify a list page builder it's possible to add extra page and list actions. For example let's assume you want to add an export button to the top of the list table:

```php
protected function modifyListBuilder(Request $request, ListPageBuilder $builder)
{
    //Add the default page actions and perform some default setup stuff
    $builder = parent::modifyListBuilder($request, $builder);
    
    //Add fields
    $builder
      ->addField('name')
    ;
    
    //Add the export action
    if ($this->hasAccess(self::ACTION_EXPORT)) {
      $builder->prependAction(new ExportPageActionElement([], [
        'route' => $request->get('_route'),
        'action' => self::ACTION_EXPORT,
      ]));
    }
    
    return $builder;
}
```

The list page builder provides the `addAction` and `prependAction` functions to easily add new page actions. A getter and setter for the page actions also exists in case you need to perform more advanced logic when adding page actions.
