<?php

namespace Bazookas\AdminBundle\Tests\Controller\AdminController;

use Bazookas\AdminBundle\Controller\Base\BaseAdminListController;
use Bazookas\AdminBundle\Entity\CMSUser;
use Bazookas\AdminBundle\Tests\TestCase\AdminTestCase;
use Symfony\Component\DomCrawler\Crawler;
use Symfony\Bundle\FrameworkBundle\Client;

class BaseCMSUserAdminControllerTest extends AdminTestCase
{

  /**
   * Test that pages load successfully when logged in as admin
   * @return array
   */
  public function getUrls() {
    $urls = [
      $this->getUrl('admin_cms_user', [
        '_locale' => $this->defaultLocale,
        '_action' => BaseAdminListController::ACTION_LIST
      ])
    ];

    $parentUrls = parent::getUrls();
    return array_merge($urls, $parentUrls);
  }

  /**
   * Test adding a new cms user as a super admin
   */
  public function testAddUserSuperAdmin() {
    list($client, $crawler) = $this->executeRequest('admin_cms_user', [
      '_locale' => $this->defaultLocale,
      '_action' => BaseAdminListController::ACTION_ADD
    ]);

    $this->addUser($crawler, $client);
  }

  /**
   * Test adding a new cms user as an admin user
   */
  public function testAddUserAdmin() {
    $adminUser = $this->referenceRepository->getReference('user-admin1');

    $client = $this->createLoggedInClient($adminUser);
    $url = $this->getUrl('admin_cms_user', [
      '_locale' => $this->defaultLocale,
      '_action' => BaseAdminListController::ACTION_ADD
    ]);

    $crawler = $client->request('GET', $url);

    $this->addUser($crawler, $client);
  }

  /**
   * Utility function to make it easier to repeat tests with different users
   * @param Crawler $crawler
   * @param Client $client
   */
  private function addUser(Crawler $crawler, Client $client) {
    $form = $crawler->selectButton('submit')->form();
    $form['cms_user_admin[name]'] = 'test1';
    $form['cms_user_admin[surname]'] = 'test1';
    $form['cms_user_admin[email]'] = 'test@bazookas.be';
    $form['cms_user_admin[username]'] = 'test';
    $form['cms_user_admin[locale]']->select($this->defaultLocale);
    $form['cms_user_admin[plainPassword][first]'] = 'test';
    $form['cms_user_admin[plainPassword][second]'] = 'test';

    $client->submit($form);

    //Assert that the form submitted successfully
    $this->assertTrue($client->getResponse()->isRedirect());

    //Assert that the item was edited
    $entityManager = $client->getContainer()->get('doctrine.orm.entity_manager');
    $entityRepo = $entityManager->getRepository(CMSUser::class);
    $entity = $entityRepo->findOneBy(['name' => 'test1']);

    $this->assertTrue(!empty($entity));
  }

  /**
   * Test editing an existing cms user as a super admin
   */
  public function testEditUser() {
    list($client, $crawler) = $this->executeRequest('admin_cms_user', [
      '_locale' => $this->defaultLocale,
      '_action' => BaseAdminListController::ACTION_EDIT,
      '_id' => 1
    ]);

    $form = $crawler->selectButton('submit')->form();
    $form['cms_user_admin[name]'] = 'test2';
    $form['cms_user_admin[surname]'] = 'test1';
    $form['cms_user_admin[email]'] = 'test@bazookas.be';
    $form['cms_user_admin[username]'] = 'test';
    $form['cms_user_admin[locale]']->select($this->defaultLocale);

    $client->submit($form);

    //Assert that the form submitted successfully
    $this->assertTrue($client->getResponse()->isRedirect());

    //Assert that the item was edited
    $entityManager = $client->getContainer()->get('doctrine.orm.entity_manager');
    $entityRepo = $entityManager->getRepository(CMSUser::class);
    $entity = $entityRepo->findOneBy(['name' => 'test2']);

    $this->assertTrue(!empty($entity));
  }

  /**
   * Test if an admin user has access to the user list page
   */
  public function testRoleAdminUser() {
    $adminUser = $this->referenceRepository->getReference('user-admin1');

    $this->testPageIsSuccessful($adminUser, [
      $this->getUrl('cms_user_admin', [
        '_locale' => $this->defaultLocale,
        '_action' => BaseAdminListController::ACTION_LIST
      ])
    ]);
  }

  /**
   * Check that normal cms users don't have access to this page (ROLE_CMS_USER)
   */
  public function testRequireRoleAdminUser() {
    $cmsUser = $this->referenceRepository->getReference('user-basic1');
    $this->testPageFails($cmsUser, [
      $this->getUrl('cms_user_admin', [
        '_locale' => $this->defaultLocale,
        '_action' => BaseAdminListController::ACTION_LIST
      ]),
      $this->getUrl('cms_user_group_admin', [
        '_locale' => $this->defaultLocale,
        '_action' => BaseAdminListController::ACTION_ADD
      ]),
      $this->getUrl('cms_user_group_admin', [
        '_locale' => $this->defaultLocale,
        '_action' => BaseAdminListController::ACTION_EDIT
      ])
    ]);
  }
}
