<?php

namespace Bazookas\APIFrameworkBundle\Formatter\Base;
use Bazookas\APIFrameworkBundle\Exception\FormatterException;
use Bazookas\CommonBundle\Utility\Utility;

class OutputFormatter
{
  //conversion types
  const TYPE_STRING = 'string';
  const TYPE_INTEGER = 'integer';

  protected $config;
  protected $globalConfig;

  public function __construct($config, $globalConfig = []) {
    $this->config = $config;
    $this->globalConfig = $globalConfig;
  }

  public function format(array &$data) {
    if (empty($data)) {
      return $data;
    }

    //Carry out field transformations
    $data = $this->applyFieldTransformations($data, $this->config);

    //Carry out global transformations
    $data = $this->applyGlobalTransformations($data, $this->globalConfig);

    return $data;
  }

  public function applyFieldTransformations(array $data, array $configuration) {
    if (empty($configuration)) {
      return $data;
    }

    foreach($configuration as $field => $transformationConfig) {
      if (Utility::isAssociative($transformationConfig)) {
        if (!isset($data[$field])) {
          continue;
        }

        try {
          $data[$field] = $this->executeTransformation($data[$field], $transformationConfig);
        } catch (FormatterException $e) {
          throw new FormatterException('Transformation error for field \''.$field.'\': ' . $e->getMessage());
        }
      } else {
        foreach ($transformationConfig as $config) {
          $data = $this->applyFieldTransformations($data, [$field => $config]);
        }
      }
    }

    return $data;
  }

  public function applyGlobalTransformations(array $data, array $configuration) {
    if (empty($configuration)) {
      return $data;
    }

    foreach($configuration as $transformationConfig) {
      if (Utility::isAssociative($transformationConfig)) {
        try {
          $data = $this->executeTransformation($data, $transformationConfig, true);
        } catch (FormatterException $e) {
          throw new FormatterException('Global transformation error: ' . $e->getMessage());
        }
      } else {
        $data = $this->applyGlobalTransformations($data, $transformationConfig);
      }
    }

    return $data;
  }

  private function executeTransformation($data, array $configuration, $isGlobal = false) {
    $transformation = isset($configuration['transformation']) ? $configuration['transformation'] : null;

    //Is a transformation defined?
    if (empty($transformation)) {
      throw new FormatterException('No transformation defined');
    }

    $params = isset($configuration['params']) ? $configuration['params'] : [];
    if (!is_array($params)) {
      $params = [$params];
    }

    if (!is_array($data) || Utility::isAssociative($data) || $isGlobal) {
      return $this->transformData($data, $params, $transformation);
    } else {
      foreach($data as $key => &$item) {
        $data[$key] = $this->transformData($item, $params, $transformation);
      }

      return $data;
    }
  }

  private function transformData($data, $params, $transformation) {
    array_unshift($params, $data);

    if (is_callable($transformation)) {
      return call_user_func_array($transformation, $params);
    } else if ($transformation instanceof OutputFormatter) {
      return $transformation->format($data);
    } else {
      throw new FormatterException('The transformation \''. print_r($transformation, true) .'\' must be a callable or an instance of \''. OutputFormatter::class .'\'.');
    }
  }

  /* ---------------------------------------------------------------------------------------------
   *                              Default field transformation functions
   * ---------------------------------------------------------------------------------------------- */
  protected function convertToType($value, $type) {
    switch ($type) {
      case self::TYPE_STRING:
        return (string) $value;
      case self::TYPE_INTEGER:
        return (int) $value;
      default:
        return $value;
    }
  }

  /**
   * @deprecated
   */
  protected function propertyToScalarArray($data, $property = 'id') {
    return $this->propertyToScalar($data, $property);
  }

  protected function propertyToScalar($element, $property = 'id') {
    return isset($element[$property]) ? $element[$property] : null;
  }

  /* ---------------------------------------------------------------------------------------------
   *                              Default global transformation functions
   * ---------------------------------------------------------------------------------------------- */

  protected function addToParent($data, $field) {
    if (!isset($data[$field])) {
      return $data;
    }

    foreach($data[$field] as $key => $item) {
      if (!isset($data[$key])) {
        $data[$key] = $item;
      } else {
        throw new \Exception('The key \''.$key.'\' already exists in \''.print_r($data, true).'\'');
      }
    }

    //remove the previous parent key
    unset($data[$field]);

    return $data;
  }

  protected function convertArrayToSingleton($data, $field) {
    if (isset($data[$field])) {
      $data[$field] = $data[$field][0];
    }

    return $data;
  }

  protected function unsetFields($data, $fields) {
    if (!is_array($fields)) {
      $fields = [$fields];
    }

    foreach($fields as $field) {
      unset($data[$field]);
    }

    return $data;
  }

  protected function replaceKeyWithValue($data, $fields) {
    if (!Utility::isAssociative($fields)) {
      throw new \Exception('Excepted association array \'[<oldKey> => <newKey>]\' got \''.print_r($fields, true).'\'');
    }

    foreach($fields as $oldKey => $newKeyIndex) {
      if (isset($data[$oldKey])) {
        $newKey = $data[$oldKey][$newKeyIndex];
        $data[$newKey] = $data[$oldKey];
        unset($data[$oldKey]);
      }
    }

    return $data;
  }

  protected function transformFieldKey($data, $param) {
    if (is_callable($param)) {
      foreach($data as $key => $value) {
        if (is_callable($param)) {
          $newKey = call_user_func($param, $key);
          $data[$newKey] = $value;
          unset($data[$key]);
        }
      }
    }

    return $data;
  }

  protected function renameKey($data, $fields) {
    if (!Utility::isAssociative($fields)) {
      throw new \Exception('Excepted association array \'[<oldKey> => <newKey>]\' got \''.print_r($fields, true).'\'');
    }

    foreach($fields as $oldKey => $newKey) {
      if (isset($data[$oldKey])) {
        $data[$newKey] = $data[$oldKey];
        unset($data[$oldKey]);
      }
    }

    return $data;
  }

  protected function usePropertyValueAsKey($data, $field) {
    if (empty($data)) {
      return $data;
    }

    if (Utility::isAssociative($data)) {
      $value = $data[$field];
      unset($data[$field]);
      return [$value => $data];
    } else {
      foreach($data as $key => $item) {
        $value = $item[$field];
        unset($item[$field]);
        $data[$value] = $item;
        unset($data[$key]);
      }

      return $data;
    }
  }
}
