##Getting started

This page will provide a brief introduction and guide to using the forgot password functionality in APIFrameworkBundle. The following sections follow a step-wise approach to implementing the forgot password feature in your API. So listen up kids!

###Creating the API controller

Lets start by creating the API controller. There is no preconfigured API controller provided as it is very trivial to create your own controller using the APIFrameworkBundle. You will therefore be responsible for creating your own route and controller. A quick start example of the api controller is provided below.

An example controller:
```php
<?php

namespace APIBundle\Controller\v1_0;

use APIBundle\Exception\ExceptionMap;
use Bazookas\APIFrameworkBundle\Controller\Base\BaseController;
use Symfony\Component\HttpFoundation\ParameterBag;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Validator\Constraints\Email;
use Symfony\Component\Validator\Constraints\NotBlank;

class AuthController extends BaseController
{
  public function forgotPasswordAction(Request $request) {
    $retrieveParams =  array(
      'Email' => [
        'constraint' => [new NotBlank(), new Email()],
        'required' => true
      ]
    );

    $extraParams = new ParameterBag(array(
      'invalidUserExceptionCode' => ExceptionMap::USER_NOT_FOUND
    ));

    return $this->handleRequest($request, $retrieveParams, 'forgotPassword', Response::HTTP_OK, $extraParams);
  }

}
```

###Creating the DataService

A DataService class has been created which contains a default implementation of the forgot password functionality. The `BaseAuthDataService` class itself can be found [here](Services/Data/Base/BaseAuthDataService). 

The class should be used by first creating your own data service and then extending the base service. An example implementation can be found below:

```php
<?php

namespace APIBundle\Services\Data;

use APIBundle\Repository\UserRepository;
use Bazookas\APIFrameworkBundle\Services\Data\Base\BaseAuthDataService;
use Symfony\Component\Routing\Generator\UrlGeneratorInterface;

class AuthDataService extends BaseAuthDataService
{
  protected $router;

  public function __construct(UserRepository $repository, $mailer, $templateEngine, $translator, $router, $fromEmail) {
    $this->repository = $repository;
    $this->router = $router;
    parent::__construct($mailer, $templateEngine, $translator, $fromEmail);
  }

  /**
   * @param $user the user!
   * @return string a url where the user can reset their password
   */
  public function getForgotPasswordLink($user) {
    return $this->router->generate('user_forgot_password', array('hash' => $user->getForgotPasswordHash()), UrlGeneratorInterface::ABSOLUTE_URL);
  }

  public function getForgotPasswordPlainTextMessage($link, $user) {
    $from = $this->translator->trans('api.email.user.forgotPassword.from', [], 'admin';

    return  $this->translator->trans('api.email.user.forgotPassword.body', [
      '%link%' => $link,
      '%firstName%' => $user->getUsername(),
      '%from%' => $from
    ], 'admin');
  }
  
  public function getForgotPasswordTemplate() {
    return 'APIBundle:email:forgotPassword.html.twig';
  }
}
```

There are a couple of important things to note in the above code sample. In order to generate a route as shown in the example above your parent service must inject the router service ("@router" argument in services.yml). Secondly you must use a User entity which implements the `ResettablePasswordInterface` interface. A base user entity has been created which already implements this interface see the BaseAppUser class in the CommonBundle. Your project specific AppUser will just need to extend this Base user. More info on the BaseAppUser class can be found below.

The data service class implements 2 abstract method which you can use to customize the forgot password feature in a project specific manner. Firstly you must implement the `getForgotPasswordLink` function the purpose of this function is to return a url to the page containing the password reset form.

####Working with the BaseAppUser and the BaseUserRepository

The BaseAppUser contains a basic implementation of an AppUser which is compatible with the forgot password feature. A companion repository for the BaseAppUser has also been created (see CommonBundle\Repository\BaseUserRepository). Your project specific repository class can directly extend it. No additional functions must be implemented in order to get the forgot password feature working (yaay!).

Note that you need to add the repository extending the BaseUserRepository to the entity annotation in your entity class for this functionality to work.

###Creating the web controller

A `BaseAuthController` has been created to render the password reset form. You should create your own controller extending this base controller as shown in the example below. The only function you will need to implement is the `getUserEntity` function which returns the class name of the user entity. NOTE: for the form to work you class must extend the BaseAppUser class. If it does not you will need create a custom form and controller in order to handle this part of the forgot password flow.

```php
<?php

namespace WebsiteBundle\Controller;

use APIBundle\Entity\AppUser;
use Bazookas\CommonBundle\Controller\Base\BaseAuthController;
use Symfony\Component\HttpFoundation\Request;

class AuthController extends BaseAuthController {

  public function getUserEntity() {
    return AppUser::class;
  }
}

```
