## Creating API Controller

This component allows you to easily create controllers for your API routes. But how does the APIController work. The APIController component consists of three sub-components. The controller itself, a validation service for validating api input and a data service to carry out any data operations.

### Creating the controller

First start of by creating a route and a controller. Creating a route will not be covered here as the symfony documentation covers it really well (but you should already know this!). In order to create an api controller you need to create a normal controller which extends the [BaseController](Controller/Base/BaseController) as shown in the example below. 

```php
<?php

namespace APIBundle\Controller\v1_0;

use Symfony\Component\HttpFoundation\Request;
use Bazookas\APIFrameworkBundle\Controller\Base\BaseController;

class MyAPIController extends BaseController
{

  public function myAction(Request $request) {
    $retrieveParams = [];
  
    return $this->handleRequest($request, $retrieveParams, 'myDataServiceFunction');
  }
}
```

This example shows the most basic API controller you can create. The API controller can do everything you expect a basic api controller to be able to do. It handles a request, passes this request and any `$params` you want to the data service function you've defined. The data service function is passed as a string to the `handleRequest` function of the Base controller (in the example above this is `myDataServiceFunction`).
 
 The base controller is able to handle any kind of API call (whether it be an HTTP GET, POST, PUT, PATCH, etc.) and is able to automatically detect the kind of api call that is being made.

### Validating input
#### Basic usage
Another important aspect to any API is the ability to to validate input. For this we can use the validation service which is built into the base controller. Using the validation service is very easy as shown in the example below.

```php
<?php

namespace APIBundle\Controller\v1_0;

use Symfony\Component\HttpFoundation\Request;
use Bazookas\APIFrameworkBundle\Controller\Base\BaseController;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\ParameterBag;
use Symfony\Component\Validator\Constraints\NotBlank;
use Symfony\Component\Validator\Constraints\Regex;
use Symfony\Component\Validator\Constraints\GreaterThan;
use Symfony\Component\Validator\Constraints\DateTime;
use Bazookas\APIFrameworkBundle\Services\Validation\ValidationService;

class MyAPIController extends BaseController
{

  public function myAction(Request $request) {
    $retrieveParams = array(
      'parameterOne' => [
        'constraint' => new NotBlank(),
        'required' => true
      ],
      'parameterTwo' => [
        'constraint' => [new NotBlank(), new Regex('/^\d+$/')],
        'required' => true
      ],
      'dateParameter' => array(
        'constraint' => [new DateTime(), new GreaterThan(
          \DateTime::createFromFormat('Y-m-d', '1900-01-01')
        )],
        'required' => true,
        'convertToType' => ValidationService::TYPE_DATETIME,
        'conversionFormat' => 'Ymd'
      ),
      'arrayParam' => [
        'constraint' =>  new All(new Collection([
            'fields' => [
              'tag' => [
                new NotBlank(),
                new Length(['max' => 100]),
              ],
              'type' => [new Choice([
                Tag::TYPE_RESTAURANT,
                Tag::TYPE_DISH,
                Tag::TYPE_OCCASION,
                Tag::TYPE_CUSTOM,
              ])],
            ],
            'required' => true
      ]))
      ]
    );

    $extraParams = new ParameterBag(array(
      'user' => $this->getUser()
    ));

    return $this->handleRequest($request, $retrieveParams, 'myDataServiceFunction', Response::HTTP_OK, $extraParams);
  }
}

```

Create an array of parameter definitions like the `$retrieveParams` array shown above. It is possible to define one or more constraints (these are symfony validator constriaints, more info [here](https://symfony.com/doc/current/reference/constraints.html)). A constraint can be passed directly as done with `parameterOne` or as an array of constraints as shown with `parameterTwo`. 

It's also possible to define if a conversion must be applied to a value before the constraint is applied. Lets take the `dateParameter' as an example. The type that the value should be converted to is defined using a validation service constant; `ValidationService::TYPE_DATETIME` in this case. An additional format can be passed to help convert the value to the given format. In this case the passed datetime value (`"19910211"`) will be converted using the `Ymd` format to create a php DateTime object. 
 
 The following date types are defined:
 
 * `DATETIME`
 * `TYPE_STRING`
 
 The format can one of the following:
 
 * a string (when dealing with the DateTime type only)
 * a php callable
 
 #### Working with exclusions and dependencies
 
 Using the validation service it's possible to define rudimentary relationships between parameters. For example let's assume we have the following:
 
 ```php
  <?php
 
 namespace APIBundle\Controller\v1_0;
 
 use Symfony\Component\HttpFoundation\Request;
 use Bazookas\APIFrameworkBundle\Controller\Base\BaseController;
 use Symfony\Component\HttpFoundation\Response;
 use Symfony\Component\HttpFoundation\ParameterBag;
 use Symfony\Component\Validator\Constraints\NotBlank;
 use Symfony\Component\Validator\Constraints\Regex;
 use Symfony\Component\Validator\Constraints\GreaterThan;
 use Symfony\Component\Validator\Constraints\DateTime;
 use Bazookas\APIFrameworkBundle\Services\Validation\ValidationService;
 
 class MyAPIController extends BaseController
 {
 
   public function myAction(Request $request) {
     $retrieveParams = array(
       'parameterOne' => [
         'constraint' => new NotBlank(),
         'required' => false,
         'dependencies' => 'parameterTwo'
       ],
       'parameterTwo' => [
         'constraint' => [new NotBlank(), new Regex('/^\d+$/')],
         'required' => false
       ],
       'dateParameter' => array(
         'constraint' => [new DateTime(), new GreaterThan(
           \DateTime::createFromFormat('Y-m-d', '1900-01-01')
         )],
         'required' => true,
         'convertToType' => ValidationService::TYPE_DATETIME,
         'conversionFormat' => 'Ymd',
         'excludes' => 'arrayParam'
       ),
       'arrayParam' => [
         'constraint' =>  new All(new Collection([
             'fields' => [
               'tag' => [
                 new NotBlank(),
                 new Length(['max' => 100]),
               ],
               'type' => [new Choice([
                 Tag::TYPE_RESTAURANT,
                 Tag::TYPE_DISH,
                 Tag::TYPE_OCCASION,
                 Tag::TYPE_CUSTOM,
               ])],
             ],
             'required' => true
       ]))
       ]
     );
 
     $extraParams = new ParameterBag(array(
       'user' => $this->getUser()
     ));
 
     return $this->handleRequest($request, $retrieveParams, 'myDataServiceFunction', Response::HTTP_OK, $extraParams);
   }
 }
 
 ```
 
 In the class above the parameter `paramterOne` as defined in `$retrieveParams` has a dependency on `parameterTwo`. What this means is that if `parameterOne` is present in the input data the validation service will ensure that `parameterTwo` becomes required. If `parameterOne` is not included in the input the dependency is not enforced (because it's not a required parameter).
 
 Similarly the `excludes` configuration can be used to state that a parameter, if required or included in the input, excludes the presence of another parameter. In the example above the `dataParamter` excludes the inclusion of the `arrayParam` because its defined as being `required` this means that the exclusion will always be enforced. If `dateParameter` is set to be optional (`required = false`) the exclusion will only be enforced if the `dateParameter` is present in the input data.
 
 #### Notes
 
Another feature of the validation service is that it will only retrieve the parameters defined in `$retrieveParams` and pass these on to the data service. All other parameters that are sent along with the api call will be ignored. The validation service will return the parameters as a ParameterBag object.

Please note that GET requests may only contain URL variables and POST/PUT/PATCH calls should not use URL variables as these parameters will not be retrieved automatically by the validation service.

In addition to automatically validating and retrieving parameters it is possible to pass along extra parameters to the data service function. Note that these extra parameters are not validated and will be passed to the data service as is. In the example above we retrieve the authenticated user and pass it along to the data service.
 

### Creating a data service

So now we have a controller and we've validated the given input. Its now time to process the input. Input processing is done using a data service. A data service is a php class that extends the [DataServiceInterface](../Services/Data/Interfaces/DataServiceInterface) interface. An example implementation is shown below.

```php
<?php

namespace APIBundle\Services\Data;

use Bazookas\APIFrameworkBundle\Services\Data\Base\BaseDataService;
use Symfony\Component\HttpFoundation\ParameterBag;

class MyDataService extends BaseDataService
{

  public function myDataServiceFunction(ParameterBag $retrieveParams, ParameterBag $extraParams) {

    // do some stuff with the parameters

    return array('data' => $data);
  }
}
```

In order to get the data service working we have to define it as a symfony service in `services.yml` as shown below. A data service is essentially any class which implements the `DataServiceInterface` and which has been defined as a symfony service with the tag `APIFrameworkBundle.Data.DataService`.

```YAML
  apibundle.dataService.MyDataService:
    class: APIBundle\Services\Data\MyDataService
    tags:
      - { name: APIFrameworkBundle.Data.DataService }
```

Now how does it all work? The api controller will loop over the pool of defined data services and find a service which implements a public function `myDataServiceFunction`. This function is then executed and two `ParameterBag` parameters are passed to the function. The first bag contains the `$retrieveParams` we defined in the API controller, the second contains the extra parameters we passed along to the `handleRequest` function.

Any array output returned by the data service function will be converted to json format and returned as an API response.

### Throwing errors

It's not always possible to return a successfull response. The BaseController therefore automatically catches and converts any exceptions implementing the `ApiExceptionInterface` a [BaseApiExceptions](Exception/Base/BaseApiException.php) has been implemented and can be used as both a guide and a base for any custom exceptions.
 
In order for exceptions to be more easily handled by API consumers it's good practice to give each exception a unique code. APIFrameworkBundle exceptions have codes defined in the [BaseExceptionMap]
(../Exception/Base/BaseExceptionMap). It is advised you make a custom ExceptionMap class for each project and define all your exception codes as constants. An example class is shown below.

```php
<?php

namespace APIBundle\Exception;

use Bazookas\APIFrameworkBundle\Exception\Base\BaseExceptionMap;

class ExceptionMap extends BaseExceptionMap
{
  const USER_EXISTS_EXCEPTION = 'user_exists';
  const USER_USERNAME_EXISTS = 'user_username_exists';
}
```
