<?php

namespace Bazookas\IntegrationsBundle\Services\Facebook;

use Bazookas\IntegrationsBundle\Exception\IntegrationsExceptionMap;
use Facebook\Facebook;
use Facebook\Exceptions\FacebookResponseException;
use Bazookas\CommonBundle\Services\Exception\ExceptionManager;
use Bazookas\IntegrationsBundle\Exception\FacebookIntegrationException;
use Bazookas\IntegrationsBundle\Exception\FacebookTokenExpiredException;
use Facebook\FacebookRequest;

class FacebookService {
  //Facebook error codes
    //TODO add general error codes

  //Authentication error codes
  const ERROR_AUTHENTICATION_APP_NOT_INSTALLED = 458;
  const ERROR_AUTHENTICATION_USER_CHECKPOINTED = 459;
  const ERROR_AUTHENTICATION_PASSWORD_CHANGED = 460;
  const ERROR_AUTHENTICATION_EXPIRED = 463;
  const ERROR_AUTHENTICATION_UNCONFIRMED_USER = 464;
  const ERROR_AUTHENTICATION_INVALID_ACCESS_TOKEN = 467;

  /**
   * @var Facebook
   */
  private $facebook;
  private $logger;
  private $exceptionManager;
  private $defaultFields;
  private $appAccessToken;
  private $idParameters;

  public function __construct(array $settings) {
    if (!empty($settings['appId']) && !empty($settings['appSecret'])) {
      $this->setCredentials($settings['appId'], $settings['appSecret']);
    }
    $this->idParameters = [
      'field' => $settings['facebookIdField'],
      'getter' => 'get' . ucfirst($settings['facebookIdField']),
    ];

    $this->defaultFields = $settings['fields'];
  }

  public function getAppId() {
    if ($this->facebook) {
      return $this->facebook->getApp()->getId();
    }
    return null;
  }

  public function setCredentials($appId, $appSecret) {
    // create a completely new facebook instance, facebook does not editing credentials
    $this->facebook = new Facebook([
      'app_id' => $appId,
      'app_secret' => $appSecret,
      'default_graph_version' => 'v2.8',
    ]);
    $this->facebook->setDefaultAccessToken("$appId|$appSecret");
  }

  /**
   * Get a user's profile
   * @param string $accessToken user or app access token, if using an app access token, userId must be provided
   * @param string $userId optional user id, defaults to 'me'
   * @param string $fields comma separated list of fields to retrieve
   * @return mixed
   */
  public function getProfile($accessToken, $userId = 'me', $fields = null) {
    $fields = $this->getDefaultFacebookFields($fields);

    $response = $this->getFacebookResponse("/$userId?fields=$fields", $accessToken);
    return $response->getGraphUser();
  }

  /**
   * @param $accessToken
   * @param $graphNode string graph node to retrieve https://developers.facebook.com/docs/graph-api/reference/
   * @param string $fields
   * @param array $parameters array of parameters to append to request
   * @return \Facebook\FacebookResponse
   */
  public function getGeneralFacebookResponse($accessToken, $graphNode = '/me', $fields = null, $parameters = null) {
    $fields = $this->getDefaultFacebookFields($fields);
    if (!empty($parameters)) {
      $parameters = '&' . http_build_query($parameters);
    }

    $response = $this->getFacebookResponse("$graphNode?fields=$fields$parameters", $accessToken);
    return $response;
  }

  public function getLongLivedToken($accessToken, $returnFullBody = false) {
    $this->throwIfNotConfigured();

    $params = array(
      'grant_type' => 'fb_exchange_token',
      'client_id' => $this->facebook->getApp()->getId(),
      'client_secret' => $this->facebook->getApp()->getSecret(),
      'fb_exchange_token' => $accessToken,
    );
    $response = $this->getFacebookResponse('/oauth/access_token?'.http_build_query($params), $accessToken);

    $responseBody = $response->getDecodedBody();
    if ($returnFullBody) {
      return $responseBody;
    }
    else {
      return $responseBody['access_token'];
    }
  }

  public function setExceptionManager(ExceptionManager $exceptionManager) {
    $this->exceptionManager = $exceptionManager;
    return $this;
  }

  /**
   * @param $endpoint
   * @param null $accessToken
   * @param null $eTag
   * @param null $graphVersion
   * @return \Facebook\FacebookResponse
   */
  private function getFacebookResponse($endpoint, $accessToken = null, $eTag = null, $graphVersion = null) {
    $this->throwIfNotConfigured();

    try {
      $response = $this->facebook->get($endpoint, $accessToken, $eTag, $graphVersion);
      return $response;
    } catch(FacebookResponseException $e) {
      $responseBody = $e->getResponse()->getDecodedBody();
      $errorCode = $responseBody['error']['code'];
      if (isset($responseBody['error']['sub_code'])) {
        $errorCode = $responseBody['error']['sub_code'];
      }

      switch($errorCode) {
        case self::ERROR_AUTHENTICATION_EXPIRED:
          $this->throwException(new FacebookTokenExpiredException('The token has expired.', IntegrationsExceptionMap::FACEBOOK_TOKEN_EXPIRED));
          break;
        default:
          $this->throwException(new FacebookIntegrationException($e->getMessage(), IntegrationsExceptionMap::FACEBOOK_API_ERROR, $e));
      }
    } catch(Facebook\Exceptions\FacebookSDKException $e) {
      $this->throwException(new FacebookIntegrationException($e->getMessage(), IntegrationsExceptionMap::FACEBOOK_INTEGRATION_EXCEPTION, $e));
    }
  }

  /**
   * execute a FacebookRequest object, used for paging
   * @param $request FacebookRequest
   * @return \Facebook\FacebookResponse
   */
  public function executeRequest(FacebookRequest $request) {
    // TODO try catch is the same as "getFacebookResponse", see if we can join this
    $this->throwIfNotConfigured();

    try {
      $response = $this->facebook->getClient()->sendRequest($request);
      return $response;
    } catch(FacebookResponseException $e) {
      $responseBody = $e->getResponse()->getDecodedBody();
      $errorCode = $responseBody['error']['code'];
      if (isset($responseBody['error']['sub_code'])) {
        $errorCode = $responseBody['error']['sub_code'];
      }

      switch($errorCode) {
        case self::ERROR_AUTHENTICATION_EXPIRED:
          $this->throwException(new FacebookTokenExpiredException('The token has expired.'));
          break;
        default:
          $this->throwException(new FacebookIntegrationException($e->getMessage(), 0, $e));
      }
    } catch(Facebook\Exceptions\FacebookSDKException $e) {
      $this->throwException(new FacebookIntegrationException($e->getMessage(), 0, $e));
    }
  }

  private function throwException(\Exception $exception) {
    if (!empty($this->exceptionManager)) {
      $exception->setIsLoggable(true);
      $this->exceptionManager->throwException($exception);
    } else {
      throw $exception;
    }
  }


  public function updateLocalProfile($user, $userData, $fieldMap = null) {
    // update profile data
    if (empty($fieldMap)) {
      $fieldMap = $this->defaultFields;
    }
    foreach ($fieldMap as $field => $setter) {
      $fieldData = $userData->getField($field);
      $method = $setter;
      if (is_array($setter)) {
        $method = $setter['method'];
      }

      if ($field === 'picture' && !empty($fieldData)) {
        $fieldData = $fieldData->getField('url');
        if (!empty($setter['imageClass'])) {
          /** @var $image \Bazookas\MediaBundle\Entity\Image */
          $image = $user->{str_replace('set', 'get', $method)}();
          if (empty($image)) {
            $image = new $setter['imageClass']();
          }
          $image->setUrl($fieldData);
          $image->setAspectRatio($setter['width'].':'.$setter['height']);
          $image->setTitle('User: '.$user->getId());
          $fieldData = $image;
        }
      }

      $user->{$method}($fieldData);
    }
  }

  private function getDefaultFacebookFields($fields) {
    if (!empty($fields)) {
      if (is_array($fields)) {
        $fields = implode(',', $fields);
      }
      return $fields;
    }

    $result = array();
    foreach ($this->defaultFields as $key => $value) {
      $field = $key;
      if ($field === 'picture') {
        if (isset($value['width'])) {
          $field .= ".width({$value['width']})";
        }
        if (isset($value['height'])) {
          $field .= ".height({$value['height']})";
        }
      }
      else if (is_array($value)) {
        if (!empty($value['fields'])) {
          $field .= '{' . implode(',', $value['fields']) . '}';
        }
      }
      $result[] = $field;
    }
    return implode(',', $result);
  }

  private function throwIfNotConfigured() {
    if (empty($this->facebook)) {
      $this->throwException(new FacebookIntegrationException(
        'Facebook credentials not set. Configure your app in config.yml or use $facebookService->setCredentials method',
        IntegrationsExceptionMap::INTEGRATION_NOT_CONFIGURED
      ));
    }
  }

  public function getIdData() {
    return [$this->idParameters['field'], $this->idParameters['getter']];
  }
}
