<?php

namespace Bazookas\MediaBundle\EventListeners;

use Doctrine\ORM\Event\LifecycleEventArgs;
use Bazookas\MediaBundle\Services\Media\Image\ImageService;
use Bazookas\MediaBundle\Entity\Image;
use Bazookas\MediaBundle\Entity\Video;
use Bazookas\MediaBundle\Entity\Interfaces\MediaEntityInterface;
use Bazookas\MediaBundle\Entity\MediaFolder;
use Bazookas\MediaBundle\Entity\Base\BaseMedia;
use Doctrine\ORM\EntityManagerInterface;
use Bazookas\MediaBundle\Exception\LinkedMediaEntityException;

class MediaRemoveListener {

  public function preRemove(LifecycleEventArgs $args)
  {
    $entity = $args->getEntity();
    $entityManager = $args->getEntityManager();

    if ($entity instanceof MediaEntityInterface) {
      $this->checkAssociatedEntities($entityManager, $entity);
    }
  }

  public function checkAssociatedEntities(EntityManagerInterface $entityManager, $entity) {
    $associatedEntities = $this->getAssociatedEntities($entityManager ,$entity);
    $entityMetaData = $entityManager->getClassMetadata(get_class($entity));

    $linkedEntities = [];
    $count = 0;
    foreach($associatedEntities as $associatedEntity) {
      //If the association is set to cascade remove then we don't care if there are still items linked to it.
      if (isset($associatedEntity['inversedBy'])) {
        $associatedMapping = $entityMetaData->getAssociationMapping($associatedEntity['inversedBy']);

        if ($associatedMapping['isCascadeRemove'] || !$associatedMapping['isOwningSide']) {
          continue;
        }
      }

      //don't query the parent relation for a media folder as being a subfolder does not impact
      //whether or not a folder should be deleted. So don't carry out the check in that case.
      if ($entity instanceof MediaFolder &&
        $associatedEntity['fieldName'] === 'children') {
        continue;
      }

      //if a media item is linked to a media folder then this does not impact whether or not the item is
      //deletable. So don't carry out the check in that case.

      if ($entity instanceof BaseMedia &&
        $associatedEntity['fieldName'] === 'mediaItems') {
        continue;
      }

      if (!$associatedEntity['isCascadeRemove']) {
        $linked = $this->queryAssociation($entityManager, $associatedEntity, $entity);

        if (!empty($linked)) {

          $linkedEntities[$associatedEntity['targetEntity']] = $linked;
          $count = $count + count($linked);
        }
      }

      if ($count > 10) {
        break;
      }
    }

    if (!empty($linkedEntities)) {
      throw new LinkedMediaEntityException('This item is linked to another entity.', 0, null, $linkedEntities);
    }
  }

  public function getAssociatedEntities(EntityManagerInterface $entityManager, $entity) {
    $metas = $entityManager->getMetadataFactory()->getAllMetadata();
    $entityClass = get_class($entity);
    $entityMetaData = $entityManager->getClassMetadata($entityClass);

    $associatedEntities = [];
    foreach ($metas as $meta) {
      $mappings = $meta->getAssociationMappings();

      foreach($mappings as $key => $mapping) {
        if ($mapping['targetEntity'] === $entityClass || $mapping['targetEntity'] === $entityMetaData->rootEntityName) {
          $associatedEntities[$mapping['sourceEntity'].$mapping['fieldName']] = $mapping;
        }
      }
    }

    return $associatedEntities;
  }

  public function queryAssociation(EntityManagerInterface $entityManager, array $associatedEntityMapping, $entity) {
    $qb = $entityManager->createQueryBuilder();
    $qb
      ->select('ae, eam')
      ->from($associatedEntityMapping['sourceEntity'], 'ae')
      ->join('ae.'.$associatedEntityMapping['fieldName'], 'eam', 'WITH', $qb->expr()->eq('eam.id', ':eamID'))
      ->setParameter('eamID', $entity->getId())
      ->setMaxResults(10)
    ;

    return $qb->getQuery()->getResult();
  }

}
