<?php

namespace Bazookas\MediaBundle\Services\Media\Image;

use Imagine\Image\ImagineInterface;
use Symfony\Bundle\FrameworkBundle\Routing\Router;
use Symfony\Component\HttpFoundation\File\UploadedFile;
use Symfony\Component\Filesystem\Filesystem;
use Bazookas\CommonBundle\Services\Util\ConnectorService;
use Liip\ImagineBundle\Imagine\Cache\CacheManager;
use Bazookas\MediaBundle\Entity\Interfaces\MediaEntityInterface;
use Symfony\Component\HttpFoundation\File\MimeType\MimeTypeExtensionGuesser;
use Bazookas\MediaBundle\Services\Media\Interfaces\MediaServiceInterface;
use Doctrine\ORM\EntityManagerInterface;
use \Bazookas\CommonBundle\Utility\Utility;
use Imagine\Image\Box;
use Bazookas\MediaBundle\Exception\InvalidMediaEntityException;

class ImageService implements MediaServiceInterface{

  const IMAGE_TARGET_MAX_WIDTH = 960;
  const IMAGE_TARGET_MAX_HEIGHT = 680;

  private $fileSystem;
  private $connector;
  private $imagineCacheManager;
  private $webDirectory;
  private $uploadsDirectory;
  private $entityManager;
  private $baseUrl;
  private $imagine;

  public function __construct(ConnectorService $connector, CacheManager $imagineCacheManager, ImagineInterface $imagine, Router $router, $mediaSettings) {
    $this->fileSystem = new Filesystem();
    $this->connector = $connector;
    $this->imagineCacheManager = $imagineCacheManager;
    $this->imagine = $imagine;

    $this->webDirectory = realpath($mediaSettings['root_directory']);
    $this->uploadsDirectory = realpath($mediaSettings['uploads_root_directory']);

    if (!$this->fileSystem->exists($mediaSettings['uploads_root_directory'])) {
      $this->fileSystem->mkdir($mediaSettings['uploads_root_directory']);
    }

    if (empty($settings['baseUrl'])) {
      $context = $router->getContext();
      $this->baseUrl =
        $context->getScheme().'://'.
        $context->getHost();
      ;
      if ($context->getScheme() === 'http') {
        if ($context->getHttpPort() != 80) {
          $this->baseUrl .= ':'.$context->getHttpPort();
        }
      }
      $this->baseUrl .= $context->getBaseUrl();
    }
    else {
      $this->baseUrl = $settings['baseUrl'];
    }
  }

  public function setEntityManager(EntityManagerInterface $entityManager) {
    $this->entityManager = $entityManager;
  }

  public function handleUpload($file, MediaEntityInterface $mediaItem, $directory = null) {
    if (empty($file)) {
      return $mediaItem;
    }

    if (!empty($directory) && !$this->fileSystem->exists($this->uploadsDirectory.'/'.$directory)) {
      $this->fileSystem->mkdir($this->uploadsDirectory.'/'.$directory);
    }

    if ($file instanceof UploadedFile) {
      $this->handleUploadedFile($file, $mediaItem, $directory);
    }

    if (is_string($file)) {
      $this->handleBase64EncodedString($file, $mediaItem, $directory);
    }

    return $mediaItem;
  }

  private function handleUploadedFile(UploadedFile $file, MediaEntityInterface $mediaItem, $directory) {
    // handle the file upload itself
    $fileName = date('Ymd').'_'.md5(uniqid()).'.'.$file->guessExtension();
    $file->move($this->uploadsDirectory.'/'.$directory, $fileName);

    $absolutePath = $this->uploadsDirectory.'/'.$directory.'/'.$fileName;

    // and finally update the media item itself
    return $this->updateMediaItem($mediaItem, $absolutePath);
  }

  private function handleBase64EncodedString($file, MediaEntityInterface $mediaItem, $directory) {
    // convert the base64 into an image
    $fileArray = explode(',', $file);

    if (count($fileArray) < 2) {
      throw new InvalidMediaEntityException('The provided image does not contain valid base64 encoding. Make sure the image string begins with \'data:<image_mime_type>;base64,<stringified data>\' for example:  \'data:image/jpeg;base64,/9j/4RTGRXhpZ...\'');
    }

    $file = $fileArray[1];

    $startPos = strpos($fileArray[0], ':')+1;
    $endPos = strpos($fileArray[0], ';', $startPos);
    $mimeType = substr($fileArray[0], $startPos, ($endPos - $startPos));

    $extensionGuesser = new MimeTypeExtensionGuesser();
    $fileType = $extensionGuesser->guess($mimeType);

    $fileName = date('Ymd').'_'.md5(uniqid()).'.'.$fileType;
    $absolutePath = $this->uploadsDirectory.'/'.$directory.'/'.$fileName;

    // write the file
    file_put_contents($absolutePath, base64_decode($file));

    // and finally update the media item itself
    return $this->updateMediaItem($mediaItem, $absolutePath);
  }

  public function downloadExternalImage($url, MediaEntityInterface $mediaItem, $directory) {
    // download the file and write it to the file system
    // TODO $file is undefined?
    $fileName = date('Ymd').'_'.md5(uniqid()).'.'.$file->guessExtension();
    $absolutePath = $this->uploadsDirectory.'/'.$directory.'/'.$fileName;

    $this->connector->downloadFile($url, '', $absolutePath);

    // update the media item entity instance
    return $this->updateMediaItem($mediaItem, $absolutePath);
  }

  private function updateMediaItem(MediaEntityInterface $mediaItem, $absolutePath) {
    $relativePath = $this->fileSystem->makePathRelative($absolutePath, $this->webDirectory);
    $this->clearImagineCache($absolutePath);

    if ($mediaItem->getPath() !== null && $this->fileSystem->exists($mediaItem->getPath())) {
      unlink($mediaItem->getPath());
    }

    //remove any trailing slashes
    if (Utility::endsWith($relativePath, '/')) {
      $relativePath = substr($relativePath, 0, strlen($relativePath) - 1);
    }
    $absoluteUrl = $this->baseUrl. '/' . $relativePath;

    $mediaItem->setPath($absolutePath);
    $mediaItem->setRelativeUrl($relativePath);
    $mediaItem->setUrl($absoluteUrl);
    $mediaItem->setAspectRatio($this->getAspectRatioAsString($absolutePath));

    return $mediaItem;
  }

  public function getUrlForPath($path) {
    return $this->imagineCacheManager->getBrowserPath($path, 'full');
  }

  private function clearImagineCache($path) {
    $this->imagineCacheManager->remove($path);
  }

  //as "width:height"
  public function getAspectRatioAsString($path) {
    $image = $this->imagine->open($path);
    $imageSize = $image->getSize();
    $imageWidth = $imageSize->getWidth();
    $imageHeight = $imageSize->getHeight();
    return $imageSize->getWidth() . ':' . $imageSize->getHeight();
  }

}
